"""Tests on source code."""

import re
import shutil
import webbrowser
from contextlib import suppress
from itertools import chain
from pathlib import Path
from subprocess import PIPE
from typing import List, Optional, Tuple

import click
from delfino.decorators import files_folders_option, pass_args
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header, run_command_example
from delfino.validation import assert_pip_package_installed

from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.utils import ensure_reports_dir


def _delete_coverage_dat_files(reports_directory: Path, test_types: List[str]):
    for test_type in [f"-{_}" for _ in test_types] + [""]:
        with suppress(FileNotFoundError):  # Use `missing_ok=True` from Python 3.8
            (reports_directory / f"coverage{test_type}.dat").unlink()


def _run_tests(
    app_context: AppContext[CorePluginConfig],
    passed_args: Tuple[str, ...],
    files_folders: Tuple[str, ...],
    name: str = "",
) -> None:
    """Execute the tests for a given test type."""
    assert_pip_package_installed("pytest")
    assert_pip_package_installed("pytest-cov")
    assert_pip_package_installed("coverage")

    plugin_config = app_context.plugin_config

    if not files_folders:
        files_folders = (plugin_config.tests_directory / name,)
    elif name and (name not in plugin_config.test_types or not plugin_config.tests_directory):
        return

    header_name = f"{name} " if name else ""
    coverage_name = f"-{name}" if name else ""

    print_header(f"️Running {header_name}tests", icon="🔎🐛")
    ensure_reports_dir(plugin_config)

    args: List[Optional[str]] = [
        "pytest",
        "--cov",
        plugin_config.sources_directory,
        "--cov-report",
        f"xml:{plugin_config.reports_directory / f'coverage{coverage_name}.xml'}",
        "--cov-branch",
        "-vv",
        *passed_args,
        *files_folders,
    ]

    if plugin_config.pytest_modules:
        args = (
            ["python"]
            + list(chain.from_iterable(("-m", module) for module in plugin_config.pytest_modules))
            + ["--module"]
            + args
        )

    run(
        list(filter(None, args)),
        on_error=OnError.ABORT,
        env_update={"COVERAGE_FILE": plugin_config.reports_directory / f"coverage{coverage_name}.dat"},
        env_update_path={"PYTHONPATH": app_context.plugin_config.sources_directory},
    )


@click.command(help="Run unit tests.")
@files_folders_option
@pass_args
@pass_plugin_app_context
def test_unit(app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[str, ...]):
    _delete_coverage_dat_files(app_context.plugin_config.reports_directory, app_context.plugin_config.test_types)
    _run_tests(app_context, passed_args, files_folders, "unit")


@click.command(help="Run integration tests.")
@files_folders_option
@pass_args
@pass_plugin_app_context
def test_integration(
    app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[str, ...]
):
    # TODO(Radek): Replace with alias?
    _delete_coverage_dat_files(app_context.plugin_config.reports_directory, app_context.plugin_config.test_types)
    _run_tests(app_context, passed_args, files_folders, "integration")


def _get_total_coverage(coverage_dat: Path) -> str:
    """Return coverage percentage, as captured in coverage dat file; e.g., returns "100%"."""
    output = run(
        "coverage report", stdout=PIPE, env_update={"COVERAGE_FILE": coverage_dat}, on_error=OnError.EXIT
    ).stdout.decode()
    match = re.search(r"TOTAL.*?([\d.]+%)", output)
    if match is None:
        raise RuntimeError(f"Regex failed on output: {output}")
    return match.group(1)


def _combined_coverage_reports(reports_directory: Path, test_types: List[str]) -> Path:
    coverage_dat_combined = reports_directory / "coverage.dat"

    coverage_files = []  # we'll make a copy because `combine` will erase them

    for test_type in test_types:
        coverage_dat = reports_directory / f"coverage-{test_type}.dat"

        if not coverage_dat.exists():
            click.secho(
                f"Could not find coverage dat file for {test_type} tests: {coverage_dat}",
                fg="yellow",
            )
        else:
            print(f"{test_type.title()} test coverage: {_get_total_coverage(coverage_dat)}")

            temp_copy = coverage_dat.with_name(coverage_dat.name.replace(".dat", "-copy.dat"))
            shutil.copy(coverage_dat, temp_copy)
            coverage_files.append(temp_copy)

    if not coverage_files and coverage_dat_combined.exists():
        click.secho(
            "Could not find coverage dat files for individual tests but combined file exists. Using this file only.",
            fg="yellow",
        )
        return coverage_dat_combined

    run(
        ["coverage", "combine", *coverage_files],
        env_update={"COVERAGE_FILE": coverage_dat_combined},
        stdout=PIPE,
        on_error=OnError.EXIT,
    )

    return coverage_dat_combined


@click.command()
@pass_plugin_app_context
def coverage_report(app_context: AppContext[CorePluginConfig], **kwargs):
    """Analyse coverage and generate a term/HTML report.

    Combines all test types.
    """
    del kwargs  # additional unused arguments passed via `click.invoke` from other commands
    assert_pip_package_installed("coverage")

    print_header("Generating coverage report", icon="📃")
    plugin_config = app_context.plugin_config
    ensure_reports_dir(plugin_config)

    coverage_dat_combined = _combined_coverage_reports(plugin_config.reports_directory, plugin_config.test_types)
    coverage_html = plugin_config.reports_directory / "coverage-report/"

    run(
        ["coverage", "html", "-d", coverage_html],
        env_update={"COVERAGE_FILE": coverage_dat_combined},
        stdout=PIPE,
        on_error=OnError.EXIT,
    )

    print(f"Total coverage: {_get_total_coverage(coverage_dat_combined)}\n")
    print(
        f"Refer to coverage report for full analysis in '{coverage_html}/index.html'\n"
        f"Or open the report in your default browser with:\n"
        f"  {run_command_example(coverage_open, app_context)}"
    )


@click.command(help="Run all tests.")
@files_folders_option
@pass_plugin_app_context
@pass_args
def test(app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[str, ...]):
    _delete_coverage_dat_files(app_context.plugin_config.reports_directory, app_context.plugin_config.test_types)
    for name in [""] if files_folders else app_context.plugin_config.test_types:
        _run_tests(app_context, passed_args, files_folders, name)


@click.command(help="Run all tests, and generate coverage report.")
@files_folders_option
@pass_plugin_app_context
@pass_args
@click.pass_context
def test_all(click_context: click.Context, **kwargs):
    del kwargs  # passed to `test`
    click_context.forward(test)
    click_context.forward(coverage_report)


@click.command(help="Open coverage results in default browser.")
@pass_plugin_app_context
def coverage_open(app_context: AppContext[CorePluginConfig]):
    report_index = app_context.plugin_config.reports_directory / "coverage-report" / "index.html"
    if not report_index.exists():
        click.secho(
            f"Could not find coverage report '{report_index}'. Ensure that the report has been built.\n"
            "Try one of the following:\n"
            f"  {run_command_example(coverage_report, app_context)}\n"
            f"or\n"
            f"  {run_command_example(test_all, app_context)}",
            fg="red",
        )

        raise click.exceptions.Exit(code=1)
    webbrowser.open(f"file:///{report_index.absolute()}")



"""Type checking on source code."""
import os

import click
from delfino.execution import OnError, run
from delfino.terminal_output import print_header
from delfino.validation import assert_pip_package_installed

_PYPI_API_TOKEN_ENV_VAR = "PYPI_API_TOKEN"


@click.command()
def upload_to_pypi():
    """Upload package to a public Pypi repository.

    Existing versions are skipped.

    Requires `PYPI_API_TOKEN` environment variable to be set to the an upload API token.
    See https://pypi.org/help/#apitoken
    """
    assert_pip_package_installed("twine")

    print_header("Uploading package to Pypi", icon="⬆")

    pypi_api_token = os.getenv("PYPI_API_TOKEN")
    assert pypi_api_token, "The 'PYPI_API_TOKEN' environment variable to authenticate with PyPI is not set."

    if not pypi_api_token.startswith("pypi-"):
        pypi_api_token = f"pypi-{pypi_api_token}"

    run(
        "twine upload --skip-existing --non-interactive dist/*",
        env_update={"TWINE_USERNAME": "__token__", "TWINE_PASSWORD": pypi_api_token},
        on_error=OnError.EXIT,
    )



import os
import re
import shlex
import subprocess
from datetime import datetime, timedelta
from typing import Optional

import click
from click import secho
from delfino.constants import ENTRY_POINT, PackageManager
from delfino.decorators import pass_app_context
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header
from delfino.utils import ArgsType
from delfino.validation import assert_package_manager_is_known, assert_pip_package_installed

from delfino_core.backports import shlex_join
from delfino_core.commands.verify_all import verify_all
from delfino_core.config import CorePluginConfig

try:
    from git import Repo
except ImportError:
    pass


def _trace(args: str) -> subprocess.CompletedProcess:
    """Print the command before execution."""
    _args: ArgsType = shlex.split(args)
    secho(shlex_join(_args), fg="white")
    return run(_args, on_error=OnError.EXIT, capture_output=True)


class Updater:
    _FILENAME: str = ""

    @staticmethod
    def _git_root():
        return (
            run(["git", "rev-parse", "--show-toplevel"], capture_output=True, on_error=OnError.EXIT)
            .stdout.decode()
            .strip()
        )

    def __init__(self):
        assert_pip_package_installed("gitpython")

        self._repo = Repo(self._git_root())
        now = datetime.utcnow()
        self._start_of_week = now - timedelta(now.isoweekday() - 1)

    def commit_and_push(self):
        commit_message = f"Dependencies rollup: {self._start_of_week.strftime('%Y-%m-%d')}"
        can_update = self._repo.commit("HEAD").message.strip() == commit_message

        if input("\033[1;33mDo you want to commit changes now? [Y/n]: \033[0m").lower() == "y":
            if can_update:  # update existing commit
                if self._repo.is_dirty():
                    _trace("git pull")
                _trace("git commit -a --amend -C HEAD")
            else:
                _trace("git add .")
                _trace(f"git commit -a -m '{commit_message}'")

        if input("\033[1;33mDo you want to push changes now? [Y/n]: \033[0m").lower() == "y":
            if can_update:
                _trace("git push --force-with-lease")
            else:
                _trace("git push -u origin HEAD")

            url = self._link_to_open_a_pull_request()
            if url:
                secho(f"\nOpen a new pull request by visiting:\n\n\t{url}\n", fg="green")

    def _link_to_open_a_pull_request(self) -> Optional[str]:
        url = self._repo.remote().url
        match = re.match("git@github.com:(.*)\\.git", url)
        if not match:
            match = re.match("https://github.com/(.*)\\.git", url)

        return f"https://github.com/{match.group(1)}/pull/new/{self._repo.active_branch}" if match else None

    def _read_dependency_file(self) -> str:
        with open(self._FILENAME, encoding="utf-8") as file:
            return file.read()

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        raise NotImplementedError

    def get_branch_name(self) -> str:
        email = self._repo.config_reader().get_value("user", "email")
        assert (
            email
        ), "Git config has not a commit email set. Please set it with:\n\tgit config --global user.email 'YOUR_EMAIL'"
        assert isinstance(email, str)

        user_name = re.sub("@.*", "", email)
        assert user_name, f"No user name could be parsed from git commit email '{email}' after removing '@.*'."
        assert isinstance(user_name, str)
        user_name = re.sub("[^a-zA-Z0-9._-]", "_", user_name)

        return f"{user_name}/dependencies_rollup_{self._start_of_week.strftime('%Y_%m_%d')}"

    @classmethod
    def _lock_and_sync(cls):
        pass

    def _show_edit_prompt_and_wait(self, *, available_updates: str):
        print("\n" + available_updates)
        input(
            f"\033[1;33mEdit {self._FILENAME} file to update any of the dependencies above ^^^\n\n"
            f"Then continue by pressing ENTER ...\033[0m"
        )

    def checkout_branch(self, branch: str):
        if str(self._repo.active_branch) == branch:
            secho(f"Branch '{branch}' already exists and active.", fg="green")
        elif branch in self._repo.branches:  # type: ignore[operator]
            secho(f"Branch '{branch}' already exists.", fg="yellow")
            _trace(f"git checkout {branch}")
        else:
            _trace("git stash")
            if self._repo.active_branch != "main":
                _trace("git checkout main")
            _trace("git pull")
            _trace(f"git checkout -b {branch}")
            self._lock_and_sync()

    def update(self, retry: bool):
        branch_name = self.get_branch_name()
        self.checkout_branch(branch_name)

        if not retry:
            while True:
                if not self.print_outdated_packages_and_lock_if_changed():
                    break

            secho("Running all tests to check the updated dependencies ... ", fg="yellow")


class PipenvUpdater(Updater):
    _FILENAME = "Pipfile"

    _SKIP_PATTERN = (
        "Skipped Update of Package (?P<package>[^:]+): (?P<installed>[^ ]+) "
        "installed.* (?P<available>[^ ]+) available.*"
    )
    _OUTDATED_PATTEN = (
        "Package '(?P<package>[^']+)' out-of-date: '(?P<installed>[^']+)' "
        "installed.* '(?P<available>[^']+)' available.*"
    )
    _VERSION_CONSTRAINT_CHARS = "=~<>"

    @classmethod
    def _lock_and_sync(cls):
        _trace("pipenv lock")
        _trace("pipenv sync -d")

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        secho("Updating packages based on version pinning. This will take a while ...", fg="yellow")
        _trace("pipenv update -d")

        secho("Checking outdated packages. This will take a while ...", fg="yellow")
        result = _trace("pipenv update --outdated")

        pipfile = self._read_dependency_file()

        available_updates = []

        for line in result.stdout.decode().split(os.linesep) + result.stderr.decode().split(os.linesep):
            match = re.match(self._SKIP_PATTERN, line) or re.match(self._OUTDATED_PATTEN, line)
            if not match:
                continue

            package = match.group("package")
            installed = match.group("installed").lstrip(self._VERSION_CONSTRAINT_CHARS)
            available = match.group("available").lstrip(self._VERSION_CONSTRAINT_CHARS)

            # Keep only packages defined in Pipfile with a different version available
            if installed != available and re.search(f'\n"?{package}"? ', pipfile):
                available_updates.append(f"{package}: {installed} -> {available}")

        if not available_updates:
            return False

        self._show_edit_prompt_and_wait(available_updates="\n".join(sorted(available_updates)) + "\n")

        return self._read_dependency_file() != pipfile


class PoetryUpdater(Updater):
    _FILENAME = "pyproject.toml"

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        secho("Updating packages based on version pinning. This will take a while ...", fg="yellow")
        _trace("poetry update")

        secho("Checking outdated packages. This will take a while ...", fg="yellow")
        result = _trace("poetry show --outdated --why --ansi")

        if not result.stdout:
            return False

        pyproject_toml = self._read_dependency_file()

        self._show_edit_prompt_and_wait(available_updates=result.stdout.decode())

        return self._read_dependency_file() != pyproject_toml


@click.command()
@click.option("--retry", default=False, show_default=True, is_flag=True, help=" Retry an update after failed tests.")
@pass_app_context(CorePluginConfig)
@click.pass_context
def dependencies_update(click_context: click.Context, app_context: AppContext, retry):
    """Manages the process of updating dependencies."""
    print_header("Updating dependencies", icon="🔄")

    assert_package_manager_is_known(app_context.package_manager)
    updater: Updater

    if app_context.package_manager == PackageManager.PIPENV:
        updater = PipenvUpdater()
    elif app_context.package_manager == PackageManager.POETRY:
        updater = PoetryUpdater()
    else:
        raise AssertionError(
            f"The '{app_context.package_manager.value}' package manager is not supported by this command."
        )

    updater.update(retry)

    try:
        click_context.invoke(verify_all)
    except Exception:
        secho(
            f"\nOne or more checks have failed. Fix any issues above and then run:\n\n\t"
            f"{ENTRY_POINT} {click_context.info_name} --retry\n",
            fg="red",
        )
        raise

    updater.commit_and_push()



import shlex
from contextlib import suppress
from pathlib import Path
from typing import Iterable


def path_is_relative_to(path1: Path, path2: Path) -> bool:
    """Backport of ``PurePath.is_relative_to`` for Python 3.7 and 3.8."""
    with suppress(ValueError):
        path1.relative_to(path2)
        return True
    return False


def shlex_join(split_command: Iterable[str]) -> str:
    """Return a shell-escaped string from *split_command*.

    Backport of ``shlex.join`` for Python 3.7
    """
    return " ".join(shlex.quote(arg) for arg in split_command)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['delfino_core', 'delfino_core.commands']

package_data = \
{'': ['*']}

install_requires = \
['delfino>=0.28.1']

extras_require = \
{'all': ['black',
         'isort',
         'pre-commit',
         'pytest',
         'coverage',
         'pytest-cov',
         'mypy',
         'pylint',
         'pycodestyle',
         'pydocstyle',
         'twine',
         'gitpython',
         'packaging',
         'psutil'],
 'build-docker': ['packaging'],
 'dependencies-update': ['gitpython'],
 'format': ['black', 'isort', 'pre-commit'],
 'lint': ['pylint', 'pycodestyle', 'pydocstyle', 'psutil'],
 'test': ['pytest', 'coverage', 'pytest-cov'],
 'typecheck': ['mypy'],
 'upload-to-pypi': ['twine'],
 'verify-all': ['black',
                'isort',
                'pre-commit',
                'pytest',
                'coverage',
                'pytest-cov',
                'mypy',
                'pylint',
                'pycodestyle',
                'pydocstyle',
                'psutil']}

entry_points = \
{'delfino.plugin': ['delfino-core = delfino_core.commands']}

setup_kwargs = {
    'name': 'delfino-core',
    'version': '3.8.1',
    'description': 'Delfino core plugin',
    'long_description': '<h1 align="center" style="border-bottom: none;"> 🔌&nbsp;&nbsp;Delfino Core&nbsp;&nbsp; 🔌</h1>\n<h3 align="center">A <a href="https://github.com/radeklat/delfino">Delfino</a> plugin with core functionality.</h3>\n\n<p align="center">\n    <a href="https://app.circleci.com/pipelines/github/radeklat/delfino-core?branch=main">\n        <img alt="CircleCI" src="https://img.shields.io/circleci/build/github/radeklat/delfino-core">\n    </a>\n    <a href="https://app.codecov.io/gh/radeklat/delfino-core/">\n        <img alt="Codecov" src="https://img.shields.io/codecov/c/github/radeklat/delfino-core">\n    </a>\n    <a href="https://github.com/radeklat/delfino-core/tags">\n        <img alt="GitHub tag (latest SemVer)" src="https://img.shields.io/github/tag/radeklat/delfino-core">\n    </a>\n    <img alt="Maintenance" src="https://img.shields.io/maintenance/yes/2022">\n    <a href="https://github.com/radeklat/delfino-core/commits/main">\n        <img alt="GitHub last commit" src="https://img.shields.io/github/last-commit/radeklat/delfino-core">\n    </a>\n    <a href="https://www.python.org/doc/versions/">\n        <img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/delfino-core">\n    </a>\n    <a href="https://pypistats.org/packages/delfino-core">\n        <img alt="Downloads" src="https://img.shields.io/pypi/dm/delfino-core">\n    </a>\n</p>\n\n# Commands\n  \n| Command               | Description                                         |\n|-----------------------|-----------------------------------------------------|\n| build-docker          | Build and push a docker image.                      |\n| coverage-open         | Open coverage results in default browser.           |\n| coverage-report       | Analyse coverage and generate a term/HTML report.   |\n| dependencies-update   | Manages the process of updating dependencies.       |\n| format                | Runs black code formatter and isort on source code. |\n| lint                  | Run linting on the entire code base.                |\n| lint-pycodestyle      | Run PEP8 checking on code.                          |\n| lint-pydocstyle       | Run docstring linting on source code.               |\n| lint-pylint           | Run pylint on code.                                 |\n| switch-python-version | Switches Python venv to a different Python version. |\n| test-all              | Run all tests, and generate coverage report.        |\n| test-integration      | Run integration tests.                              |\n| test-unit             | Run unit tests.                                     |\n| typecheck             | Run type checking on source code.                   |\n| upload-to-pypi        | Upload package to a public Pypi repository.         |\n| verify-all            | Runs all verification commands.                     |\n\n# Installation\n\n- pip: `pip install delfino-core`\n- Poetry: `poetry add -D delfino-core`\n- Pipenv: `pipenv install -d delfino-core`\n\n## Optional dependencies\n\nEach project may use different sub-set of [commands](#commands). Therefore, dependencies of all commands are optional and checked only when the command is executed.\n\nUsing `[all]` installs all the [optional dependencies](https://setuptools.pypa.io/en/latest/userguide/dependency_management.html#optional-dependencies) used by all the commands. If you want only a sub-set of those dependencies, there are finer-grained groups available:\n\n- For individual commands (matches the command names):\n  - `upload-to-pypi`\n  - `build-docker`\n  - `typecheck`\n  - `format`\n  - `dependencies-update`\n- For groups of commands:\n  - `test` - for testing and coverage commands\n  - `lint` - for all the linting commands\n- For groups of groups:\n  - `verify-all` - same as `[typecheck,format,test,lint]`\n  - `all` - all optional packages\n\n# Configuration\n\nDelfino doesn\'t load any plugins by default. To enable this plugin, add the following config into `pyproject.toml`:\n\n```toml\n[tool.delfino.plugins.delfino-core]\n\n```\n\n## Plugin configuration\n\nThis plugin has several options. All the values are optional and defaults are shown below: \n\n```toml\n[tool.delfino.plugins.delfino-core]\n# Source files - may have different rules than tests (usually stricter)\nsources_directory = "src"\n\n# Test files\ntests_directory = "tests"\n\n# Where to store reports generated by various tools\nreports_directory = "reports"\n\n# Types of tests you have nested under the `tests_directory`. Will be executed in given order.\ntest_types = ["unit", "integration"]\n\n# One or more module to wrap `pytest` in, executing it as `python -m <MODULE> pytest ...`\npytest_modules = []\n\n# Commands to run as a quality gate in given order.\nverify_commands = ["format", "lint", "typecheck", "test-all"]\n\n# Do not install pre-commit if this is set to true.\ndisable_pre_commit = false\n```\n\n## Commands configuration\n\nSeveral commands have their own configuration as well:\n\n```toml\n[tool.delfino.plugins.delfino-core.dockerhub]\n# User name for logging in into dockerhub\nusername = ""\n\n# Platforms to build with dockerx\nbuild_for_platforms = [\n    "linux/arm/v7",\n    "linux/arm64",\n    "linux/amd64",\n]\n\n[tool.delfino.plugins.delfino-core.typecheck]\n# One or more directories where type hint will be required. By default they are optional.\nstrict_directories = []  \n```\n\n# Usage\n\nRun `delfino --help`.\n\n# Development\n\nTo develop against editable `delfino` sources:\n\n1. Make sure `delfino` sources are next to this plugin:\n    ```shell\n    cd ..\n    git clone https://github.com/radeklat/delfino.git\n    ```\n2. Install `delfino` as editable package:\n    ```shell\n    pip install -e ../delfino\n    ```\n   Note that poetry will reset this to the released package when you install/update anything.\n',
    'author': 'Radek Lát',
    'author_email': 'radek.lat@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/radeklat/delfino-core',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



from pathlib import Path
from subprocess import PIPE

import click
from delfino.constants import PackageManager
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header
from delfino.validation import assert_package_manager_is_known, pip_package_installed

from delfino_core.config import CorePluginConfig, pass_plugin_app_context


@click.command()
@click.argument("version", type=str)
@pass_plugin_app_context
def switch_python_version(app_context: AppContext[CorePluginConfig], version: str):
    """Switches Python venv to a different Python version.

    - VERSION: Desired Python version. You can use only MAJOR.MINOR (for example 3.6).

    Use this to test the sub-packages with a different Python version. CI pipeline always
    checks all supported versions automatically.

    Notes:
        This task calls `deactivate` as a precaution for cases when the task is called
        from an active virtual environment.
    """
    print_header(f"Switching to Python {version}", icon="🐍")

    pyenv_python_version = None
    python_versions = sorted(
        (_ for _ in run("pyenv versions --bare", stdout=PIPE, on_error=OnError.EXIT).stdout.decode().split("\n") if _),
        key=lambda value: list(map(int, value.split("."))),  # sort numerically
    )

    assert Path(".venv").exists(), "Folder '.venv'."

    for python_version in python_versions:
        if python_version.startswith(version):
            pyenv_python_version = python_version
            click.secho(f"✔ Found pyenv Python version '{pyenv_python_version}'.", fg="green")

    if not pyenv_python_version:
        available_python_versions = ", ".join(f"'{_}'" for _ in python_versions)
        click.secho(f"❌ No pyenv Python version matching Python {version} found.", fg="red")
        print(
            f"Available versions: {available_python_versions}.\n"
            f"See all installable versions with:\n"
            f"\tpyenv install --list\n"
            f"and install it with:\n"
            f"\tpyenv install <PYTHON_VERSION>",
        )
        raise click.Abort()

    package_manager = app_context.package_manager
    assert_package_manager_is_known(package_manager)
    if package_manager == PackageManager.POETRY:
        install_command = "poetry install --no-root"
    elif package_manager == PackageManager.PIPENV:
        install_command = "pipenv install -d --deploy"
    else:
        click.secho("No compatible package manager detected. Skipping package installation.", fg="red", err=True)
        raise click.Abort()

    click.secho("ℹ Removing current virtualenv ...", fg="blue")
    run("git clean -fxd .venv", stdout=PIPE, stderr=PIPE, on_error=OnError.EXIT)

    click.secho(f"ℹ Switching to Python {pyenv_python_version} ...", fg="blue")
    run(["pyenv", "local", pyenv_python_version], stdout=PIPE, stderr=PIPE, on_error=OnError.EXIT)

    click.secho(f"✔ Detected {package_manager.value.capitalize()} package manager.", fg="green")

    if not pip_package_installed(package_manager.value, sub_process=True):
        click.secho(f"⚠ {package_manager.value.capitalize()} is not installed. Installing ...", fg="yellow")
        run(f"pip install {package_manager.value}", stdout=PIPE, stderr=PIPE, on_error=OnError.EXIT)
    else:
        click.secho(f"✔ {package_manager.value.capitalize()} is installed.", fg="green")

    run(install_command, on_error=OnError.EXIT)
